from __future__ import annotations
import typing
from solana.publickey import PublicKey
from spl.token.constants import TOKEN_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
from ..program_id import MANGO_PROGRAM_ID


class StubOracleCloseAccounts(typing.TypedDict):
    group: PublicKey
    admin: PublicKey
    oracle: PublicKey
    sol_destination: PublicKey


def stub_oracle_close(
    accounts: StubOracleCloseAccounts,
    program_id: PublicKey = MANGO_PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["sol_destination"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\\\x89-\x03-<u\xe0"
    encoded_args = b""
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import MANGO_PROGRAM_ID


class PerpLiqBasePositionArgs(typing.TypedDict):
    max_base_transfer: int


layout = borsh.CStruct("max_base_transfer" / borsh.I64)


class PerpLiqBasePositionAccounts(typing.TypedDict):
    group: PublicKey
    perp_market: PublicKey
    oracle: PublicKey
    liqor: PublicKey
    liqor_owner: PublicKey
    liqee: PublicKey


def perp_liq_base_position(
    args: PerpLiqBasePositionArgs,
    accounts: PerpLiqBasePositionAccounts,
    program_id: PublicKey = MANGO_PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["oracle"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["liqor"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["liqor_owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["liqee"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xa8L\xc9uH5Ak"
    encoded_args = layout.build(
        {
            "max_base_transfer": args["max_base_transfer"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from ..program_id import MANGO_PROGRAM_ID


class PerpLiqForceCancelOrdersArgs(typing.TypedDict):
    limit: int


layout = borsh.CStruct("limit" / borsh.U8)


class PerpLiqForceCancelOrdersAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    perp_market: PublicKey
    bids: PublicKey
    asks: PublicKey


def perp_liq_force_cancel_orders(
    args: PerpLiqForceCancelOrdersArgs,
    accounts: PerpLiqForceCancelOrdersAccounts,
    program_id: PublicKey = MANGO_PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["bids"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["asks"], is_signer=False, is_writable=True),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"m\xcb\xba\x10\xe9[\x01\x8d"
    encoded_args = layout.build(
        {
            "limit": args["limit"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



import typing
from dataclasses import dataclass
from solana.publickey import PublicKey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import MANGO_PROGRAM_ID


class Serum3MarketIndexReservationJSON(typing.TypedDict):
    group: str
    market_index: int
    reserved: list[int]


@dataclass
class Serum3MarketIndexReservation:
    discriminator: typing.ClassVar = b"\xf6\x10\xc6d\xefpx5"
    layout: typing.ClassVar = borsh.CStruct(
        "group" / BorshPubkey, "market_index" / borsh.U16, "reserved" / borsh.U8[38]
    )
    group: PublicKey
    market_index: int
    reserved: list[int]

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: PublicKey,
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = MANGO_PROGRAM_ID,
    ) -> typing.Optional["Serum3MarketIndexReservation"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id.to_solders():
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[PublicKey],
        commitment: typing.Optional[Commitment] = None,
        program_id: PublicKey = MANGO_PROGRAM_ID,
    ) -> typing.List[typing.Optional["Serum3MarketIndexReservation"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["Serum3MarketIndexReservation"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "Serum3MarketIndexReservation":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = Serum3MarketIndexReservation.layout.parse(
            data[ACCOUNT_DISCRIMINATOR_SIZE:]
        )
        return cls(
            group=dec.group,
            market_index=dec.market_index,
            reserved=dec.reserved,
        )

    def to_json(self) -> Serum3MarketIndexReservationJSON:
        return {
            "group": str(self.group),
            "market_index": self.market_index,
            "reserved": self.reserved,
        }

    @classmethod
    def from_json(
        cls, obj: Serum3MarketIndexReservationJSON
    ) -> "Serum3MarketIndexReservation":
        return cls(
            group=PublicKey(obj["group"]),
            market_index=obj["market_index"],
            reserved=obj["reserved"],
        )



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from spl.token.constants import TOKEN_PROGRAM_ID
from solana.transaction import TransactionInstruction, AccountMeta
from ..program_id import MANGO_PROGRAM_ID


class PerpCloseMarketAccounts(typing.TypedDict):
    group: PublicKey
    admin: PublicKey
    perp_market: PublicKey
    bids: PublicKey
    asks: PublicKey
    event_queue: PublicKey
    sol_destination: PublicKey


def perp_close_market(
    accounts: PerpCloseMarketAccounts,
    program_id: PublicKey = MANGO_PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["admin"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["bids"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["asks"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["event_queue"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["sol_destination"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xe2\xee\xbb\x11\xa0\x99\xfe\xa0"
    encoded_args = b""
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
from ..program_id import MANGO_PROGRAM_ID


class PerpDeactivatePositionAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    owner: PublicKey
    perp_market: PublicKey


def perp_deactivate_position(
    accounts: PerpDeactivatePositionAccounts,
    program_id: PublicKey = MANGO_PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["perp_market"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x8cT\xf3\xf9\x1d\x94\x12\x1d"
    encoded_args = b""
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from __future__ import annotations
import typing
from solana.publickey import PublicKey
from solana.transaction import TransactionInstruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import MANGO_PROGRAM_ID


class Serum3CancelOrderArgs(typing.TypedDict):
    side: types.serum3_side.Serum3SideKind
    order_id: int


layout = borsh.CStruct("side" / types.serum3_side.layout, "order_id" / borsh.U128)


class Serum3CancelOrderAccounts(typing.TypedDict):
    group: PublicKey
    account: PublicKey
    owner: PublicKey
    open_orders: PublicKey
    serum_market: PublicKey
    serum_program: PublicKey
    serum_market_external: PublicKey
    market_bids: PublicKey
    market_asks: PublicKey
    market_event_queue: PublicKey


def serum3_cancel_order(
    args: Serum3CancelOrderArgs,
    accounts: Serum3CancelOrderAccounts,
    program_id: PublicKey = MANGO_PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> TransactionInstruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["group"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["owner"], is_signer=True, is_writable=False),
        AccountMeta(pubkey=accounts["open_orders"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["serum_market"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["serum_program"], is_signer=False, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["serum_market_external"], is_signer=False, is_writable=True
        ),
        AccountMeta(pubkey=accounts["market_bids"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["market_asks"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["market_event_queue"], is_signer=False, is_writable=True
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"~rm\x16\xc6\x0c|K"
    encoded_args = layout.build(
        {
            "side": args["side"].to_encodable(),
            "order_id": args["order_id"],
        }
    )
    data = identifier + encoded_args
    return TransactionInstruction(keys, program_id, data)



from setuptools import setup, find_packages


setup(
    name='mango-explorer-v4',
    version='0.1',
    license='MIT',
    author="Mango Markets",
    author_email='support@mango.markets',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/blockworks-foundation/mango-explorer-v4'
)


import typing
from anchorpy.error import ProgramError


class SomeError(ProgramError):
    def __init__(self) -> None:
        super().__init__(6000, "")

    code = 6000
    name = "SomeError"
    msg = ""


class NotImplementedError(ProgramError):
    def __init__(self) -> None:
        super().__init__(6001, "")

    code = 6001
    name = "NotImplementedError"
    msg = ""


class MathError(ProgramError):
    def __init__(self) -> None:
        super().__init__(6002, "checked math error")

    code = 6002
    name = "MathError"
    msg = "checked math error"


class UnexpectedOracle(ProgramError):
    def __init__(self) -> None:
        super().__init__(6003, "")

    code = 6003
    name = "UnexpectedOracle"
    msg = ""


class UnknownOracleType(ProgramError):
    def __init__(self) -> None:
        super().__init__(6004, "oracle type cannot be determined")

    code = 6004
    name = "UnknownOracleType"
    msg = "oracle type cannot be determined"


class InvalidFlashLoanTargetCpiProgram(ProgramError):
    def __init__(self) -> None:
        super().__init__(6005, "")

    code = 6005
    name = "InvalidFlashLoanTargetCpiProgram"
    msg = ""


class HealthMustBePositive(ProgramError):
    def __init__(self) -> None:
        super().__init__(6006, "health must be positive")

    code = 6006
    name = "HealthMustBePositive"
    msg = "health must be positive"


class HealthMustBePositiveOrIncrease(ProgramError):
    def __init__(self) -> None:
        super().__init__(6007, "health must be positive or increase")

    code = 6007
    name = "HealthMustBePositiveOrIncrease"
    msg = "health must be positive or increase"


class HealthMustBeNegative(ProgramError):
    def __init__(self) -> None:
        super().__init__(6008, "health must be negative")

    code = 6008
    name = "HealthMustBeNegative"
    msg = "health must be negative"


class IsBankrupt(ProgramError):
    def __init__(self) -> None:
        super().__init__(6009, "the account is bankrupt")

    code = 6009
    name = "IsBankrupt"
    msg = "the account is bankrupt"


class IsNotBankrupt(ProgramError):
    def __init__(self) -> None:
        super().__init__(6010, "the account is not bankrupt")

    code = 6010
    name = "IsNotBankrupt"
    msg = "the account is not bankrupt"


class NoFreeTokenPositionIndex(ProgramError):
    def __init__(self) -> None:
        super().__init__(6011, "no free token position index")

    code = 6011
    name = "NoFreeTokenPositionIndex"
    msg = "no free token position index"


class NoFreeSerum3OpenOrdersIndex(ProgramError):
    def __init__(self) -> None:
        super().__init__(6012, "no free serum3 open orders index")

    code = 6012
    name = "NoFreeSerum3OpenOrdersIndex"
    msg = "no free serum3 open orders index"


class NoFreePerpPositionIndex(ProgramError):
    def __init__(self) -> None:
        super().__init__(6013, "no free perp position index")

    code = 6013
    name = "NoFreePerpPositionIndex"
    msg = "no free perp position index"


class Serum3OpenOrdersExistAlready(ProgramError):
    def __init__(self) -> None:
        super().__init__(6014, "serum3 open orders exist already")

    code = 6014
    name = "Serum3OpenOrdersExistAlready"
    msg = "serum3 open orders exist already"


class InsufficentBankVaultFunds(ProgramError):
    def __init__(self) -> None:
        super().__init__(6015, "bank vault has insufficent funds")

    code = 6015
    name = "InsufficentBankVaultFunds"
    msg = "bank vault has insufficent funds"


class BeingLiquidated(ProgramError):
    def __init__(self) -> None:
        super().__init__(6016, "account is currently being liquidated")

    code = 6016
    name = "BeingLiquidated"
    msg = "account is currently being liquidated"


class InvalidBank(ProgramError):
    def __init__(self) -> None:
        super().__init__(6017, "invalid bank")

    code = 6017
    name = "InvalidBank"
    msg = "invalid bank"


class ProfitabilityMismatch(ProgramError):
    def __init__(self) -> None:
        super().__init__(6018, "account profitability is mismatched")

    code = 6018
    name = "ProfitabilityMismatch"
    msg = "account profitability is mismatched"


class CannotSettleWithSelf(ProgramError):
    def __init__(self) -> None:
        super().__init__(6019, "cannot settle with self")

    code = 6019
    name = "CannotSettleWithSelf"
    msg = "cannot settle with self"


class PerpPositionDoesNotExist(ProgramError):
    def __init__(self) -> None:
        super().__init__(6020, "perp position does not exist")

    code = 6020
    name = "PerpPositionDoesNotExist"
    msg = "perp position does not exist"


class MaxSettleAmountMustBeGreaterThanZero(ProgramError):
    def __init__(self) -> None:
        super().__init__(6021, "max settle amount must be greater than zero")

    code = 6021
    name = "MaxSettleAmountMustBeGreaterThanZero"
    msg = "max settle amount must be greater than zero"


class HasOpenPerpOrders(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6022, "the perp position has open orders or unprocessed fill events"
        )

    code = 6022
    name = "HasOpenPerpOrders"
    msg = "the perp position has open orders or unprocessed fill events"


class OracleConfidence(ProgramError):
    def __init__(self) -> None:
        super().__init__(6023, "an oracle does not reach the confidence threshold")

    code = 6023
    name = "OracleConfidence"
    msg = "an oracle does not reach the confidence threshold"


class OracleStale(ProgramError):
    def __init__(self) -> None:
        super().__init__(6024, "an oracle is stale")

    code = 6024
    name = "OracleStale"
    msg = "an oracle is stale"


class SettlementAmountMustBePositive(ProgramError):
    def __init__(self) -> None:
        super().__init__(6025, "settlement amount must always be positive")

    code = 6025
    name = "SettlementAmountMustBePositive"
    msg = "settlement amount must always be positive"


class BankBorrowLimitReached(ProgramError):
    def __init__(self) -> None:
        super().__init__(6026, "bank utilization has reached limit")

    code = 6026
    name = "BankBorrowLimitReached"
    msg = "bank utilization has reached limit"


class BankNetBorrowsLimitReached(ProgramError):
    def __init__(self) -> None:
        super().__init__(
            6027,
            "bank net borrows has reached limit - this is an intermittent error - the limit will reset regularly",
        )

    code = 6027
    name = "BankNetBorrowsLimitReached"
    msg = "bank net borrows has reached limit - this is an intermittent error - the limit will reset regularly"


class TokenPositionDoesNotExist(ProgramError):
    def __init__(self) -> None:
        super().__init__(6028, "token position does not exist")

    code = 6028
    name = "TokenPositionDoesNotExist"
    msg = "token position does not exist"


CustomError = typing.Union[
    SomeError,
    NotImplementedError,
    MathError,
    UnexpectedOracle,
    UnknownOracleType,
    InvalidFlashLoanTargetCpiProgram,
    HealthMustBePositive,
    HealthMustBePositiveOrIncrease,
    HealthMustBeNegative,
    IsBankrupt,
    IsNotBankrupt,
    NoFreeTokenPositionIndex,
    NoFreeSerum3OpenOrdersIndex,
    NoFreePerpPositionIndex,
    Serum3OpenOrdersExistAlready,
    InsufficentBankVaultFunds,
    BeingLiquidated,
    InvalidBank,
    ProfitabilityMismatch,
    CannotSettleWithSelf,
    PerpPositionDoesNotExist,
    MaxSettleAmountMustBeGreaterThanZero,
    HasOpenPerpOrders,
    OracleConfidence,
    OracleStale,
    SettlementAmountMustBePositive,
    BankBorrowLimitReached,
    BankNetBorrowsLimitReached,
    TokenPositionDoesNotExist,
]
CUSTOM_ERROR_MAP: dict[int, CustomError] = {
    6000: SomeError(),
    6001: NotImplementedError(),
    6002: MathError(),
    6003: UnexpectedOracle(),
    6004: UnknownOracleType(),
    6005: InvalidFlashLoanTargetCpiProgram(),
    6006: HealthMustBePositive(),
    6007: HealthMustBePositiveOrIncrease(),
    6008: HealthMustBeNegative(),
    6009: IsBankrupt(),
    6010: IsNotBankrupt(),
    6011: NoFreeTokenPositionIndex(),
    6012: NoFreeSerum3OpenOrdersIndex(),
    6013: NoFreePerpPositionIndex(),
    6014: Serum3OpenOrdersExistAlready(),
    6015: InsufficentBankVaultFunds(),
    6016: BeingLiquidated(),
    6017: InvalidBank(),
    6018: ProfitabilityMismatch(),
    6019: CannotSettleWithSelf(),
    6020: PerpPositionDoesNotExist(),
    6021: MaxSettleAmountMustBeGreaterThanZero(),
    6022: HasOpenPerpOrders(),
    6023: OracleConfidence(),
    6024: OracleStale(),
    6025: SettlementAmountMustBePositive(),
    6026: BankBorrowLimitReached(),
    6027: BankNetBorrowsLimitReached(),
    6028: TokenPositionDoesNotExist(),
}


def from_code(code: int) -> typing.Optional[CustomError]:
    maybe_err = CUSTOM_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err
